/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
// code in this file is ported/derived from from Davic McAllister's freely licensed Particle System

#include "Actions.h"
#include "Domains.h"

#include <glm/geometric.hpp>

#include <limits>

namespace Particle { namespace Actions {

/* ----------------------------------------------------------------------------------------------------------------------------- */

Gravity::Gravity(vector const& v) {
    g = v;
}
void Gravity::operator()(time dt, iterator it) {
    it->vel += g * dt;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

void Move::operator()(time dt, iterator it) {
    it->pos += it->vel * dt;
    it->age += dt;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

template<> Bounce<Domains::Disc>::Bounce(vector::value_type friction, vector::value_type _resilience, vector::value_type _cutoff, Domains::Disc&& d) : domain(d){
    oneMinusFriction = 1.f - friction;
    resilience = _resilience;
    cutoff = _cutoff;
}
template<> void Bounce<Domains::Disc>::operator()(time dt, iterator it) {
    auto& m = *it;

    // See if particle's current and next positions cross plane.
    // If not, couldn't bounce, so keep going.
    vector pnext(m.pos + m.vel * dt);

    // p2 stores the plane normal (the a,b,c of the plane eqn).
    // Old and new distances: dist(p,plane) = n * p + d
    // radius1 stores -n*p, which is d. radius1Sqr stores d.
    float distold = glm::dot(m.pos, domain.n) + domain.d;
    float distnew = glm::dot(pnext, domain.n) + domain.d;

    // Opposite signs if product < 0
    // Is there a faster way to do this?
    if(distold * distnew >= 0)
        return;

    // Find position at the crossing point by parameterizing
    // p(t) = pos + vel * t
    // Solve dist(p(t),plane) = 0 e.g.
    // n * p(t) + D = 0 ->
    // n * p + t (n * v) + D = 0 ->
    // t = -(n * p + D) / (n * v)
    // Could factor n*v into distnew = distold + n*v and save a bit.
    // Safe since n*v != 0 assured by quick rejection test.
    // This calc is indep. of dt because we have established that it
    // will hit before dt. We just want to know when.
    float nv = glm::dot(domain.n, m.vel);
    float t = -distold / nv;

    // Actual intersection point p(t) = pos + vel t
    vector phit = m.pos + m.vel * t;

    // Offset from origin in plane, phit - origin
    vector offset = phit - domain.p; 
    auto r = glm::length(offset); 
    if(r > domain.r1 || r < domain.r2)
        return;

    // Compute tangential and normal components of velocity
    vector vn = nv * domain.n; // Normal  Vn = (V.N)N
    vector vt = m.vel - vn;    // Tangent Vt = V - Vn

    // Compute new velocity heading out -- don't apply friction if tangential velocity < cutoff
    if(glm::length(vt) <= cutoff)
        m.vel = vt - vn * resilience;
    else
        m.vel = vt * oneMinusFriction - vn * resilience;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

template<> SinkPosition<Domains::Plane>::SinkPosition(bool _kill_inside, Domains::Plane&& d) : domain(d) {
    kill_inside = _kill_inside;
}
template<> void SinkPosition<Domains::Plane>::operator()(time dt, iterator it) {
    if (domain.isIn(it->pos) ^ kill_inside) {
        it->age = -999999;//std::numeric_limits<decltype(it->age)>::infinity();
        it->clr = { 255,255,0,255 };
    }
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

}}
