/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#include "Domains.h"

#include <glm/vec4.hpp>
#include <glm/gtc/random.hpp>
#include <glm/gtx/component_wise.hpp>

#include <algorithm>

namespace Particle { namespace Domains {

/* ----------------------------------------------------------------------------------------------------------------------------- */

Constant::Constant(vector::value_type const& aK) {
    k = aK;
}
void Constant::emit(vector::value_type& v) { 
    v = k; 
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Color::Color(color const& cr1, color const& cr2) {
    c1 = cr1;
    c2 = cr2;
}
void Color::emit(color& v) { 
    v = color(glm::trunc(glm::linearRand(glm::vec4(c1), glm::vec4(c2))));
}
/* ----------------------------------------------------------------------------------------------------------------------------- */

Point::Point(vector const& pt) {
    p = pt;
}
void Point::emit(vector& v) {
    v = p;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Line::Line(vector const& pt1, vector const& pt2) {
    p1 = pt1;
    p2 = pt2;
}
void Line::emit(vector& v) {
    v = glm::linearRand(p1, p2);
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Triangle::Triangle(vector const& pt1, vector const& pt2, vector const& pt3) {
    // A triangle is represented by a plane equation constrained by a triangle
    p = pt1;
    u = pt2 - pt1;
    v = pt3 - pt1;
    n = glm::normalize(glm::cross(u, v));
    d = -glm::dot(p, n);
}
void Triangle::emit(vector& v) {
    // pick a random number in [0-1]
    glm::vec2 r = glm::linearRand(glm::vec2(0), glm::vec2(1));

    // pick a point inside the triangle
    if (glm::compAdd(r) < 1.f)
        v = p + u*r[0] + v*r[1];
    else 
        v = p + u*(1.f-r[0]) + v*(1.f-r[1]);
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Plane::Plane(vector const& pt1, vector const& pt2, vector const& pt3) {
    // A plane is represented by a plane equaiton
    p = pt1;
    u = pt2 - pt1;
    v = pt3 - pt1;
    n = glm::normalize(glm::cross(u, v));
    d = -glm::dot(p, n);
}
void Plane::emit(vector& v) {
    // pick 2 random numbers between [0-1]
    glm::vec2 r = glm::linearRand(glm::vec2(0), glm::vec2(1));

    // pick a random point on the plane
    v = p + u*r[0] + v*r[1];
}
bool Plane::isIn(vector const& pt) const {    
    // Distance from plane = n * p + d; inside is the positive half-space.
    return glm::dot(n, pt) + d < 0; 
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Sphere::Sphere(vector const& p, vector::value_type r1, vector::value_type r2) {
    // A sphere is represented by a point and two radii
    this->p  = p;
    this->r1 = std::max(r1, r2);
    this->r2 = std::min(r1, r2);
}
void Sphere::emit(vector& v) {
    // pick a random point inside the sphere
    v = p + glm::sphericalRand(r1-r2) + r2;
}
bool Sphere::isIn(vector const& pt) const {
    // determine if the pt is inside the sphere
    auto r = glm::length(pt - p);
    return r2 <= r && r <= r1;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Box::Box(vector const& pt1, vector const & pt2) {
    // A box is represented by a min and max points
    pmin = glm::min(pt1, pt2);
    pmax = glm::max(pt1, pt2);
}
void Box::emit(vector& v) {
    // pick a random point in the box
    v = glm::linearRand(pmin, pmax);
}
bool Box::isIn(vector const& v) const
{
    // determin if pt is inside the box
    return 
        pmin.x <= v.x && v.x <= pmax.x &&
        pmin.y <= v.y && v.y <= pmax.y &&
        pmin.z <= v.z && v.z <= pmax.z;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Disc::Disc(vector const& pt, vector const& nr, vector::value_type rd1, vector::value_type rd2) {
    // A disc is represented by a plane equation and two radii
    p = pt;
    n = glm::normalize(nr);
    r1 = std::max(rd1, rd2);
    r2 = std::min(rd1, rd2);
    d = -glm::dot(p, n);
}
void Disc::emit(vector& v) {
    // pick a random point on the disc
    v = p + vector(glm::diskRand(r1-r2),0) + r2;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

Cylinder::Cylinder(vector const& pt1, vector const& pt2, vector::value_type rd1, vector::value_type rd2) {
    // A cyclinder is represented by a point, a vector to the other end point, and two radiii
    p1 = pt1;
    p2 = pt2 - pt1;
    r1 = std::max(rd1, rd2);
    r2 = std::min(rd1, rd2);
    auto n = glm::normalize(p2);

    // calculate two basis vectors of the plane containing the point
    vector basis(1.0f, 0.0f, 0.0f); 
    if (std::fabs(glm::dot(basis, n)) > 0.999)  basis = vector(0.0f, 1.0f, 0.0f);
    u = glm::normalize(basis - n * glm::dot(basis, n));
    v = glm::cross(n, u);
}
void Cylinder::emit(vector& o)  {
#if 1
    // there's a hole in the random number generator, we need to add a litte more to 2*PI to fill it
    vector::value_type a = glm::linearRand(0.f, 6.283185307179586476925286766559f + 0.025f); 
    auto c = glm::vec2(cos(a), sin(a)) * (r1-r2 + r2);
#else
    auto c = glm::circularRand(r1-r2 + r2);
#endif
    // pick a random point on the disc
    o = p1 + p2 * glm::linearRand(0.f,1.f) + u*c.x + v*c.y;
}

}}
