/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#include "Particles.h"
#include "Actions.h"
#include "partition.h"

#include <algorithm>
#include <vector>
#include <memory>

#include <omp.h>

namespace Particle {

// The Particle System PIMPL
struct System::Data 
{
    iterator then = nullptr, now  = nullptr;
    size_t alives = 0;

    std::unique_ptr<Source>              source;    
    std::vector<std::unique_ptr<Action>> actions;
    std::unique_ptr<Renderer>            renderer;

    double t = 0; size_t count = 0;
};

// Constructor
System::System(
    Source*             && source, 
    std::vector<Action*>&& actions, 
    Renderer*           && renderer) 
: 
    _(*new Data)
{
    _.source.reset(source);
    for(auto&& a : actions) _.actions.emplace_back(a);
    _.renderer.reset(renderer);

    _.then = _.now = reinterpret_cast<iterator>((*_.renderer).allocate(sizeof(*_.now) * nParticles));
}

// Destructor
System::~System()
{
    delete &_;
}

// Run the particle system for one time step 'dt' seconds in length
bool System::frame(time dt)
{
    // pointers to the active pool
    iterator begin = _.now;
    iterator iter  = begin + _.alives;
    iterator end   = begin + nParticles;

    // run all the actions, decrease alives as particles die
    #pragma omp parallel for schedule(static, 1024)
    for (int i = 0; i < int(_.alives); ++i) { // 'i': index variable in OpenMP 'for' statement must have signed integral type
        begin[i] = _.then[i]; for (auto& action : _.actions) 
            (*action)(dt, begin + i);
    }

    // parition the pool, reconcile the emitted with the timed-out 
#if _OPENMP
    iter = begin + __gnu_parallel::__parallel_partition(begin, end, [](auto& v) {return v.age >= 0;}, omp_get_num_procs(), 1024);
#else
    iter = std::partition(begin, end, [](auto& v) { return v.age>=0;} );
#endif

    // run the source on the unused porition of the pool, increases alives as particles are emitted
    iter = (*_.source)(dt, iter, end);

    // update the alives count
    _.alives = iter - begin;

    // render it
    bool quit = false; _.then = _.now; _.now = reinterpret_cast<iterator>((*_.renderer)(dt, begin, iter, quit));

    // Report performance snapshot
    _.t += dt; _.count += _.alives; if (_.t > 10.f) {
        fprintf(stderr, "dt: %1.8f -- alives: %zu / %u  -- particles/sec: %1.6f\n", dt, _.alives, nParticles, (_.count/_.t)/1000000.);
        _.count = 0; _.t = 0; 
    };

    return quit;
}


}
