/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#pragma once
#include <functional>
#include <memory>
#include <chrono>

#include <glm/vec3.hpp>
#include <glm/vec4.hpp>

#include <vector>

namespace Particle {

// time units for particle calculations
using time   = float;
using vector = glm::vec3;
using color  = glm::tvec4<uint8_t, glm::packed>;

// core i5/i7 cache line size: 64 bytes

__declspec(align(32))
struct AOS {
    vector pos;  // 12 bytes; position<x,y,z> -- 16-byte aligned
    color  clr;  // 4  bytes; color 
    vector vel;  // 12 bytes; velocity<x,y,z> -- 16 byte aligned
    float  age;  // 4  bytes; age
};               // 32 bytes; total -- 2 particles per cache-line
static_assert(32==sizeof(AOS), "32 != sizeof(AOS)");

// core i7 cache: L2=256Kb, L3=8Mb
// core i5 cache: L2=256Kb, L3=6Mb
#ifdef _DEBUG
static constexpr int nParticles = 10000;
#else
static constexpr int nParticles = ((8)*(1024*1024) - (1024)*(4)) /sizeof(AOS);
#endif

using item      = AOS;
using iterator  = item* __restrict;
using reference = item& __restrict;

// Components of a particle system
struct Source      { virtual iterator operator()(time, iterator, iterator)        = 0; virtual ~Source     () {} };
struct Initializer { virtual void     operator()(time, iterator)                  = 0; virtual ~Initializer() {} };
struct Action      { virtual void     operator()(time, iterator)                  = 0; virtual ~Action     () {} };
struct Renderer    { virtual void*    operator()(time, iterator, iterator, bool&) = 0; virtual ~Renderer   () {} 
                     virtual void*    allocate(size_t)                            = 0; };

// A system has all the components to run a collection of particles
struct System
{
    System(
        Source*             &&, 
        std::vector<Action*>&&, 
        Renderer*           &&
    );
    ~System();

    bool frame(time dt);

    struct Data; Data& _;
};

}
