/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#include "Renderers.h"

#include <Windows.h>

#include <glad/glad.h>

#include <glfw/glfw3.h>

#include <glm/mat4x4.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <OVR_CAPI_GL.h>
#include <OVR_CAPI_Audio.h>

#include <array>

namespace Particle { namespace Renderers { 

struct Oculus::Data
{
    struct {
        ovrSession              session              = {};
        ovrHmdDesc              hmdDesc              = {};
        ovrTextureSwapChainDesc textureSwapChainDesc = {};
        ovrTextureSwapChain     textureSwapChain     = {};
        struct {
            ovrSizei            fovTextureSize       = {};
        } eyes[2];
    } oculus;

    struct {
        GLFWwindow* win     = nullptr;
        GLuint      vShader = 0;
        GLuint      fShader = 0;
        GLuint      pgm     = 0;
        GLuint      vao     = 0;
        GLuint      vbo     = 0;
        GLuint      mvp     = 0;
        GLuint      fbo     = 0;
        GLuint      depth   = 0;
        long long   frame   = 0;
        time        time    = 0;
    } draw;

    struct {
        GLubyte*    ptr     = nullptr;
        size_t      stride  = 0;
    } memory;

    struct {
        size_t idx = 0;
        struct Item { GLbyte* ptr; GLsync sync; };
        std::array<Item, 3> items = {};
        GLbyte*&  ptr(int i = 0) { return items[next(i)].ptr; }
        GLsync&  sync(int i = 0) { return items[next(i)].sync; }
        size_t   next(int i = 1) { return (idx + offset(i)) % items.size(); }
        size_t offset(int i) { return i<0 ? items.size() + i : i; }
        void  advance() { idx = next(); }
    } pipeline;
};

// vertex shader program for point-cloud visualizer
static std::string vShader = std::string("") +
    "#version 450\n"
    "\n"
    "layout(location = 0) in vec3  in_Position;\n"
    "layout(location = 1) in vec4  in_Color;\n"
    "layout(location = 2) in vec3  in_Velocity;\n"
    "layout(location = 3) in float in_Age;\n"
    "\n"
    "uniform mat4 MVP;\n"
    "\n"
    "out vec4 v_color;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "    v_color = in_Color;\n"
    "    gl_Position = MVP *  vec4(in_Position,1);\n"
    "}\n"
    ;

// fragment shader program for point-cloud visualizer
static std::string fShader = std::string("") +
    "#version 330\n"
    "\n"
    "in vec4 v_color;\n"
    "\n"
    "layout(location = 0) out vec4 o_FragColor;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "    o_FragColor = v_color;\n"
    "}\n"
    ;

// Debug build performs OpenGL error checking, Release does not
#ifdef _DEBUG
#define GLCHK { \
    GLenum error = glGetError(); if (error != GL_NO_ERROR) { \
        auto  clamped = std::min(unsigned(error - GL_INVALID_ENUM), 4U); \
        char* errors[] = { "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION", "Unknown" }; \
        fprintf_s(stderr, "GL error at %s:%d: %s\n", __FILE__, __LINE__, errors[clamped]); \
        __debugbreak(); \
    } \
}
#else
#define GLCHK
#define __debugbreak() {}
#endif

// Static function to compile an OpenGL shader, check and report errors
static GLuint compileShader(const std::string& src, GLenum type)
{
    GLuint shader = glCreateShader(type);                                                                                   GLCHK;
    const GLchar* str = src.c_str();  glShaderSource(shader, 1, &str, NULL);                                                GLCHK;
    glCompileShader(shader);                                                                                                GLCHK;
    GLint status; glGetShaderiv(shader, GL_COMPILE_STATUS, &status); if (GL_FALSE == status) {
        GLCHK;
        GLint sz; glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &sz);                                                           GLCHK;
        std::vector<GLchar> v(sz); glGetShaderInfoLog(shader, sz, &sz, &v[0]);                                              GLCHK;
        const char* msg = &v[0]; fputs(msg, stderr); __debugbreak();
        glDeleteShader(shader);                                                                                             GLCHK;
        return 0;
    }
    return shader;
}

// Static function to link a set of OpenGL shaders into a program, check and report errors
static GLuint createProgram(std::initializer_list<GLuint> shaders)
{
    GLuint program = glCreateProgram();                                                                                     GLCHK;
    for (auto shader : shaders) glAttachShader(program, shader);                                                            GLCHK;
    glLinkProgram(program);                                                                                                 GLCHK;
    GLint status; glGetProgramiv(program, GL_LINK_STATUS, &status); if (GL_FALSE == status) {
        GLCHK;
        GLint sz; glGetProgramiv(program, GL_INFO_LOG_LENGTH, &sz);                                                         GLCHK;
        std::vector<GLchar> v(sz); glGetProgramInfoLog(program, sz, &sz, &v[0]);                                            GLCHK;
        const char* msg = &v[0]; fputs(msg, stderr); __debugbreak();
        glDeleteProgram(program);                                                                                           GLCHK;
        for (auto shader : shaders) glDeleteShader(shader);                                                                 GLCHK;
        return 0;
    }
    for (auto shader : shaders) glDetachShader(program, shader);                                                            GLCHK;
    return program;
}

// Static function to adjust the viewport based on the window's size
static void framebufferSizeCallback(GLFWwindow* win, int w, int h) {
    auto& _ = *reinterpret_cast<Oculus::Data*>(glfwGetWindowUserPointer(win));
    glViewport(0, 0, w, h);                                                                                                 GLCHK;
}

// Static function to be called to process OpenGL debug messages
static void APIENTRY debugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const void* userParam) {
    if ((type == GL_DEBUG_TYPE_ERROR) || (type == GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR))
        fprintf(stderr, "OpenGL Debug: %s\n", message);
}

// Constructor
Oculus::Oculus() : _(*new Data)
{
    // initialize Oculus
    if (ovrSuccess != ovr_Initialize(nullptr)) {
        fputs("YIKES! Could not ovr_Initialize(nullptr).\n", stderr);
        std::exit(-1);
    }        
    ovrGraphicsLuid luid; if (ovrSuccess != ovr_Create(&_.oculus.session, &luid)) {
        fputs("YIKES! Could not ovr_Create().\n", stderr);
        std::exit(-1);
    }

    // initialize windows system
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_DECORATED, GLFW_FALSE);
#   ifdef _DEBUG
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
#   endif

    // create window and opengl context, configure glad
    _.draw.win = glfwCreateWindow(800, 600, "Particles", nullptr, nullptr); if (!_.draw.win) {
        fputs("YIKES! Could not open window with GLFW.\n", stderr);
        std::exit(-1);
    }
    glfwSetWindowPos(_.draw.win, 5, 34);
    glfwSetWindowUserPointer(_.draw.win, &_);
    glfwMakeContextCurrent(_.draw.win);
    glfwSetFramebufferSizeCallback(_.draw.win, framebufferSizeCallback);
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
    }
    framebufferSizeCallback(_.draw.win, 800, 600);
#ifdef _DEBUG
    glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);                                                                                      GLCHK;
    glEnable(GL_DEBUG_OUTPUT);                                                                                                  GLCHK;
    glDebugMessageCallback(debugCallback, nullptr);                                                                             GLCHK;
#   endif
    glfwSwapInterval(0);

    // create the Oculus render targets
    _.oculus.hmdDesc = ovr_GetHmdDesc(_.oculus.session);    
    _.oculus.eyes[ovrEye_Left ].fovTextureSize = ovr_GetFovTextureSize(_.oculus.session, ovrEye_Left, _.oculus.hmdDesc.DefaultEyeFov[ovrEye_Left ], 1);
    _.oculus.eyes[ovrEye_Right].fovTextureSize = ovr_GetFovTextureSize(_.oculus.session, ovrEye_Left, _.oculus.hmdDesc.DefaultEyeFov[ovrEye_Right], 1);
    _.oculus.textureSwapChainDesc.Type         = ovrTexture_2D;
    _.oculus.textureSwapChainDesc.ArraySize    = 1;
    _.oculus.textureSwapChainDesc.Format       = OVR_FORMAT_R8G8B8A8_UNORM_SRGB;
    _.oculus.textureSwapChainDesc.Width        =          _.oculus.eyes[ovrEye_Left].fovTextureSize.w + _.oculus.eyes[ovrEye_Right].fovTextureSize.w;
    _.oculus.textureSwapChainDesc.Height       = std::max(_.oculus.eyes[ovrEye_Left].fovTextureSize.h,  _.oculus.eyes[ovrEye_Right].fovTextureSize.h);
    _.oculus.textureSwapChainDesc.MipLevels    = 1;
    _.oculus.textureSwapChainDesc.SampleCount  = 1;
    _.oculus.textureSwapChainDesc.StaticImage  = ovrFalse;
    if (ovrSuccess != ovr_CreateTextureSwapChainGL(_.oculus.session, &_.oculus.textureSwapChainDesc, &_.oculus.textureSwapChain)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_CreateTextureSwapChainGL() = \"%s\"\n", info.ErrorString);
        std::exit(-1);
    }
    int swapChainLength; if (ovrSuccess != ovr_GetTextureSwapChainLength(_.oculus.session, _.oculus.textureSwapChain, &swapChainLength)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_GetTextureSwapChainLength() = \"%s\"\n", info.ErrorString);
        std::exit(-1);
    }
    for (int i = 0; i < swapChainLength; ++i) {
        GLuint tex; if (ovrSuccess != ovr_GetTextureSwapChainBufferGL(_.oculus.session, _.oculus.textureSwapChain, 0, &tex)) {
            ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
            fprintf(stderr, "YIKES! Could not ovr_GetTextureSwapChainBufferGL() = \"%s\"\n", info.ErrorString);
            std::exit(-1);
        }
        glBindTexture(GL_TEXTURE_2D, tex);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);                                                       GLCHK;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);                                                       GLCHK;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);                                                    GLCHK;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);                                                    GLCHK;
    }

    // create the fbo
    glGenRenderbuffers(1, &_.draw.depth);                                                                                       GLCHK;
    glBindRenderbuffer(GL_RENDERBUFFER, _.draw.depth);                                                                          GLCHK;
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH24_STENCIL8, _.oculus.textureSwapChainDesc.Width, _.oculus.textureSwapChainDesc.Height); GLCHK;
    glGenFramebuffers(1, &_.draw.fbo);                                                                                          GLCHK;

    // compile and link the shaders into a program
    _.draw.vShader = compileShader(vShader, GL_VERTEX_SHADER);
    _.draw.fShader = compileShader(fShader, GL_FRAGMENT_SHADER);
    _.draw.pgm     = createProgram({ _.draw.vShader, _.draw.fShader });
    _.draw.mvp     = glGetUniformLocation(_.draw.pgm, "MVP");                                                                   GLCHK;

    // set global state
    glCreateVertexArrays(1, &_.draw.vao);                                                                                       GLCHK;
    glBindVertexArray(_.draw.vao);                                                                                              GLCHK;
    glEnableVertexAttribArray(0);                                                                                               GLCHK;
    glEnableVertexAttribArray(1);                                                                                               GLCHK;
    glEnableVertexAttribArray(2);                                                                                               GLCHK;
    glEnableVertexAttribArray(3);                                                                                               GLCHK;
    glEnable(GL_DEPTH_TEST);                                                                                                    GLCHK;
    glEnable(GL_FRAMEBUFFER_SRGB);                                                                                              GLCHK;
    glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);                                                                                    GLCHK;
    glPointSize(1);                                                                                                             GLCHK;

    // start playing the fountain sound
    // note:  this sound file was created by recording the sound of the Oculus lancher program
    PlaySound(L"oculus.wav", NULL, SND_FILENAME | SND_ASYNC | SND_LOOP);
}

Oculus::~Oculus()
{
    glfwTerminate();
    ovr_DestroyTextureSwapChain(_.oculus.session, _.oculus.textureSwapChain);
    ovr_Destroy(_.oculus.session);
    ovr_Shutdown();
    delete &_;
}

void* Oculus::allocate(size_t size)
{
    // remember the stride in the memory pipeline
    _.memory.stride = size; size *= _.pipeline.items.size();

    // create the buffer
    GLbitfield flags = GL_MAP_READ_BIT | GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT;
    glGenBuffers(1, &_.draw.vbo);                                                               GLCHK;
    glBindBuffer(GL_ARRAY_BUFFER, _.draw.vbo);                                                  GLCHK;
    glBufferStorage(GL_ARRAY_BUFFER, size, nullptr, flags);                                     GLCHK;
    _.memory.ptr = (GLubyte*) glMapBufferRange(GL_ARRAY_BUFFER, 0, size, flags);                GLCHK;
    glBindBuffer(GL_ARRAY_BUFFER, 0);                                                           GLCHK;

    // create the pipeline
    GLbyte* ptr = (GLbyte*)_.memory.ptr; for(int i=0; i<int(_.pipeline.items.size()); ++i, ptr+=_.memory.stride)
        _.pipeline.ptr(i) = ptr;

    // return pointer to first pipeline item
    std::memset(_.memory.ptr, 0xFF, size);
    return _.pipeline.ptr();
}

void* Oculus::operator()(time dt, iterator begin, iterator end, bool& quit)
{
    // check the status of the Oculus session
    ovrSessionStatus sessionStatus = {}; if (ovrSuccess != ovr_GetSessionStatus(_.oculus.session, &sessionStatus)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_GetSessionStatus() = \"%s\"\n", info.ErrorString);
    }
    if (sessionStatus.ShouldQuit) {
        quit = true; return _.pipeline.ptr();;
    }
    if (!sessionStatus.IsVisible) {
        return _.pipeline.ptr();
    }
    if (sessionStatus.ShouldRecenter && ovrSuccess != ovr_RecenterTrackingOrigin(_.oculus.session)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_RecenterTrackingOrigin() = \"%s\"\n", info.ErrorString);
    }

    // get the oculus rendering information for both eyes
    auto displayMidpointSeconds     = ovr_GetPredictedDisplayTime(_.oculus.session, 0);
    auto sensorSampleTime           = ovr_GetTimeInSeconds();
    auto hmdState                   = ovr_GetTrackingState(_.oculus.session, displayMidpointSeconds, ovrTrue);
    ovrEyeRenderDesc renderDesc[] = { ovr_GetRenderDesc(_.oculus.session, ovrEye_Left,  _.oculus.hmdDesc.DefaultEyeFov[ovrEye_Left]),
                                      ovr_GetRenderDesc(_.oculus.session, ovrEye_Right, _.oculus.hmdDesc.DefaultEyeFov[ovrEye_Right]) };
    ovrVector3f hmdToEyeOffset [] = { renderDesc[ovrEye_Left ].HmdToEyeOffset,
                                      renderDesc[ovrEye_Right].HmdToEyeOffset };
    ovrLayerEyeFov layer = {};
    layer.Header.Type                = ovrLayerType_EyeFov;
    layer.Header.Flags               = ovrLayerFlag_TextureOriginAtBottomLeft;
    layer.SensorSampleTime           = sensorSampleTime;
    layer.ColorTexture[ovrEye_Left ] = _.oculus.textureSwapChain;
    layer.ColorTexture[ovrEye_Right] = _.oculus.textureSwapChain;
    layer.Fov         [ovrEye_Left ] = renderDesc[ovrEye_Left ].Fov;
    layer.Fov         [ovrEye_Right] = renderDesc[ovrEye_Right].Fov;
    layer.Viewport    [ovrEye_Left ] = ovrRecti{ 0,                                           0, _.oculus.eyes[ovrEye_Left ].fovTextureSize.w, _.oculus.textureSwapChainDesc.Height };
    layer.Viewport    [ovrEye_Right] = ovrRecti{ _.oculus.eyes[ovrEye_Left].fovTextureSize.w, 0, _.oculus.eyes[ovrEye_Right].fovTextureSize.w, _.oculus.textureSwapChainDesc.Height };
    ovr_CalcEyePoses(hmdState.HeadPose.ThePose, hmdToEyeOffset, layer.RenderPose);

    // attach the color surface to the fbo and make it active
    int idx; if (ovrSuccess != ovr_GetTextureSwapChainCurrentIndex(_.oculus.session, _.oculus.textureSwapChain, &idx)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_GetTextureSwapChainLength() = \"%s\"\n", info.ErrorString);
    }
    GLuint tex; if (ovrSuccess != ovr_GetTextureSwapChainBufferGL(_.oculus.session, _.oculus.textureSwapChain, idx, &tex)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_GetTextureSwapChainBufferGL() = \"%s\"\n", info.ErrorString);
    }
    glBindFramebuffer(GL_FRAMEBUFFER, _.draw.fbo);                                                                              GLCHK;
    glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, tex, 0);                                   GLCHK;
    glFramebufferRenderbuffer(GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, _.draw.depth);                         GLCHK;
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER)) {
        fputs("YIKES! GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER)\n", stderr);
    }

    // clear the fbo
    glClearColor(0, 0, 0, 1);                                                                                                   GLCHK;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);                                                                         GLCHK;
    glEnable(GL_FRAMEBUFFER_SRGB);                                                                                              GLCHK;

    // render the eyes ...
    for (int eye = 0; eye < ovrEye_Count; ++eye)
    {
        // set the viewport
        auto& vp = layer.Viewport[eye];
        glViewport(vp.Pos.x, vp.Pos.y, vp.Size.w, vp.Size.h);                                                                   GLCHK;

        // bind the vao and configure the vertex array pointers -- changes every frame since offset changes every frame
        glBindVertexArray(_.draw.vao);                                                                                          GLCHK;
        glBindBuffer(GL_ARRAY_BUFFER, _.draw.vbo);                                                                              GLCHK;
        glVertexAttribPointer(0, sizeof(begin->pos)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->pos - _.memory.ptr)); GLCHK;
        glVertexAttribPointer(1, sizeof(begin->clr)/sizeof(GLbyte),  GL_UNSIGNED_BYTE,  GL_TRUE,  sizeof(*begin), (GLvoid*)((GLubyte*)&begin->clr - _.memory.ptr)); GLCHK;
        glVertexAttribPointer(2, sizeof(begin->vel)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->vel - _.memory.ptr)); GLCHK;
        glVertexAttribPointer(3, sizeof(begin->age)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->age - _.memory.ptr)); GLCHK;

        // calculate the projection matrix
        auto ovrProj = ovrMatrix4f_Projection(layer.Fov[eye], 0.01f, 1000.f, ovrProjection_ClipRangeOpenGL);
        glm::mat4 proj = glm::transpose(glm::make_mat4(&ovrProj.M[0][0]));

        // calculate the view matrix
        glm::mat4 orientation = glm::mat4_cast(glm::make_quat((float*)&layer.RenderPose[eye].Orientation));
        glm::mat4 translation = glm::translate(glm::make_vec3((float*)&layer.RenderPose[eye].Position));
        glm::mat4 view        = glm::inverse(translation * orientation);

        // calculate the model matrix
        glm::mat4 model = glm::lookAt(glm::vec3{ 0,-1.5,1.0 }, { 0,0,-0.3 }, { 0,0,1 });

        // draw the current frame
        auto mvp = proj * view * model;
        glUseProgram(_.draw.pgm);                                                                                               GLCHK;
        glUniformMatrix4fv(_.draw.mvp, 1, GL_FALSE, glm::value_ptr(mvp));                                                       GLCHK;
        glDrawArrays(GL_POINTS, 0, GLsizei(end-begin));                                                                         GLCHK;
    }

    // submit the eyes to Oculus
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0);                                          GLCHK;
    glFramebufferRenderbuffer(GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, 0);                                    GLCHK;
    glBindFramebuffer(GL_FRAMEBUFFER, 0);                                                                                       GLCHK;
    if (ovrSuccess != ovr_CommitTextureSwapChain(_.oculus.session, _.oculus.textureSwapChain)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_CommitTextureSwapChain() = \"%s\"\n", info.ErrorString);
    }
    ovrLayerHeader* header = &layer.Header; if (ovrSuccess != ovr_SubmitFrame(_.oculus.session, 0, nullptr, &header, 1)) {
        ovrErrorInfo info; ovr_GetLastErrorInfo(&info);
        fprintf(stderr, "YIKES! Could not ovr_SubmitFrame() = \"%s\"\n", info.ErrorString);
    }

    // let Windows have some time
    if ((_.draw.time += dt) > 0.5)
        glfwPollEvents();

    // mark the end of the current frame
    auto& cur = _.pipeline.sync(); if (cur) { glDeleteSync(cur);                                GLCHK; }
    cur = glFenceSync(GL_SYNC_GPU_COMMANDS_COMPLETE, 0);                                        GLCHK;

    // advance to next frame
    _.pipeline.advance();

    // wait for the next frame to be free -- wait for up to 1 second
    auto& nxt = _.pipeline.sync(); if (nxt) {
        GLenum rc = glClientWaitSync(nxt, 0, 0); if (GL_ALREADY_SIGNALED != rc) {
            for (size_t i = 0; GL_ALREADY_SIGNALED != rc && GL_TIMEOUT_EXPIRED != rc && i<1 * 1000; ++i) {
                rc = glClientWaitSync(nxt, GL_SYNC_FLUSH_COMMANDS_BIT, 1000 * 1000);              GLCHK;
            }
            fputs("OpenGL::glClientWaitSync(), waited\n", stderr);
        }
    }

    // return the pointer to the next frame
    return _.pipeline.ptr();
}

} }
