/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#include "Renderers.h"


#define NOMINMAX
#include <glad/glad.h>
#include <glfw/glfw3.h>
#include <glm/mat4x4.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <algorithm>
#include <array>
#include <iostream>
namespace Particle { namespace Renderers { 

/* ----------------------------------------------------------------------------------------------------------------------------- */

struct OpenGL::Data 
{      
    struct {
        GLFWwindow* win     = nullptr;
        GLuint      vShader = 0;
        GLuint      fShader = 0;
        GLuint      pgm     = 0;
        GLuint      vao     = 0;
        GLuint      vbo     = 0;
        GLuint      mvp     = 0;
    } draw;

    struct {
        time        time    = 0;
        glm::mat4x4 view    = {};
        glm::mat4x4 projection = {};
    } camera;

    struct {
        GLubyte*    ptr     = nullptr;
        size_t      stride  = 0;
    } memory;

    struct {
        size_t idx = 0;
        struct Item { GLbyte* ptr; GLsync sync; };
        std::array<Item, 3> items = {};
        GLbyte*&  ptr(int i=0) { return items[next(i)].ptr;             }
        GLsync&  sync(int i=0) { return items[next(i)].sync;            }
        size_t   next(int i=1) { return (idx+offset(i)) % items.size(); }
        size_t offset(int i  ) { return i<0 ? items.size() + i : i;     }
        void  advance()        { idx = next();                          }
    } pipeline;
};

// vertex shader program for point-cloud visualizer
static std::string vShader = std::string("") +
    "#version 450\n"
    "\n"
    "layout(location = 0) in vec3  in_Position;\n"
    "layout(location = 1) in vec4  in_Color;\n"
    "layout(location = 2) in vec3  in_Velocity;\n"
    "layout(location = 3) in float in_Age;\n"
    "\n"
    "uniform mat4 MVP;\n"
    "\n"
    "out vec4 v_color;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "    v_color = in_Color;\n"
    "    gl_Position = MVP *  vec4(in_Position,1);\n"
"}\n"
    ;

// fragment shader program for point-cloud visualizer
static std::string fShader = std::string("") +
    "#version 330\n"
    "\n"
    "in vec4 v_color;\n"
    "\n"
    "layout(location = 0) out vec4 o_FragColor;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "    o_FragColor = v_color;\n"
    "}\n"
    ;

// Debug build performs OpenGL error checking, Release does not
#ifdef _DEBUG
#define GLCHK { \
    GLenum error = glGetError(); if (error != GL_NO_ERROR) { \
        auto  clamped = std::min(unsigned(error - GL_INVALID_ENUM), 4U); \
        char* errors[] = { "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION", "Unknown" }; \
        fprintf_s(stderr, "GL error at %s:%d: %s\n", __FILE__, __LINE__, errors[clamped]); \
        __debugbreak(); \
    } \
}
#else
#define GLCHK
#define __debugbreak() {}
#endif

// Static function to compile an OpenGL shader, check and report errors
static GLuint compileShader(const std::string& src, GLenum type)
{
    GLuint shader = glCreateShader(type);                                                        GLCHK;
    const GLchar* str = src.c_str();  glShaderSource(shader, 1, &str, NULL);                    GLCHK;
    glCompileShader(shader);                                                                    GLCHK;
    GLint status; glGetShaderiv(shader, GL_COMPILE_STATUS, &status); if (GL_FALSE == status) {  GLCHK;
        GLint sz; glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &sz);                               GLCHK;
        std::vector<GLchar> v(sz); glGetShaderInfoLog(shader, sz, &sz, &v[0]);                  GLCHK;
        const char* msg = &v[0]; fputs(msg, stderr); __debugbreak();
        glDeleteShader(shader);                                                                 GLCHK;
        return 0;
    }
    return shader;
}

// Static function to link a set of OpenGL shaders into a program, check and report errors
static GLuint createProgram(std::initializer_list<GLuint> shaders)
{
    GLuint program = glCreateProgram();                                                         GLCHK;
    for (auto shader : shaders) glAttachShader(program, shader);                                GLCHK;
    glLinkProgram(program);                                                                     GLCHK;
    GLint status; glGetProgramiv(program, GL_LINK_STATUS, &status); if (GL_FALSE == status) {   GLCHK;
        GLint sz; glGetProgramiv(program, GL_INFO_LOG_LENGTH, &sz);                             GLCHK;
        std::vector<GLchar> v(sz); glGetProgramInfoLog(program, sz, &sz, &v[0]);                GLCHK;
        const char* msg = &v[0]; fputs(msg, stderr); __debugbreak();
        glDeleteProgram(program);                                                               GLCHK;
        for (auto shader : shaders) glDeleteShader(shader);                                     GLCHK;
        return 0;
    }
    for (auto shader : shaders) glDetachShader(program, shader);                                GLCHK;
    return program;
}

// Static function to adjust the viewport based on the window's size
static void framebufferSizeCallback(GLFWwindow* win, int w, int h) { 
    OpenGL::Data& _ = *reinterpret_cast<OpenGL::Data*>(glfwGetWindowUserPointer(win));
    _.camera.projection = glm::perspective(glm::radians(45.f), float(w)/h, 0.01f, 1000.f);
    glViewport(0, 0, w, h);                                                                     GLCHK;
}

// Static function to be called to process OpenGL debug messages
static void APIENTRY debugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const void* userParam) {
//    if ((type == GL_DEBUG_TYPE_ERROR) || (type == GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR))
        fprintf(stderr, "OpenGL Debug: %s\n", message);
}

// Constructor
OpenGL::OpenGL() : _(*new Data)
{
    // initialize windows system
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
#   ifdef _DEBUG
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
#   endif

    // create window and opengl context, configure glad
    _.draw.win = glfwCreateWindow(800, 600, "Particles", nullptr, nullptr); if (!_.draw.win) {
        fputs("YIKES! Could not open window with GLFW.\n", stderr);
        std::exit(2);
    }
    glfwSetWindowPos(_.draw.win, 5,34);
    glfwSetWindowUserPointer(_.draw.win, &_);
    glfwMakeContextCurrent(_.draw.win);
    glfwSetFramebufferSizeCallback(_.draw.win, framebufferSizeCallback);
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
        fputs("YIKES! Could not gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)", stderr);
        std::exit(2);
    }
    framebufferSizeCallback(_.draw.win, 800, 600);
#ifdef _DEBUG
    glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
    glEnable(GL_DEBUG_OUTPUT);
    glDebugMessageCallback(debugCallback, nullptr);
#   endif
    glfwSwapInterval(0);

    // compile and link the shaders into a program
    _.draw.vShader = compileShader(vShader, GL_VERTEX_SHADER);
    _.draw.fShader = compileShader(fShader, GL_FRAGMENT_SHADER);
    _.draw.pgm = createProgram({ _.draw.vShader, _.draw.fShader });
    _.draw.mvp = glGetUniformLocation(_.draw.pgm, "MVP");                                       GLCHK;

    // set global state
    glCreateVertexArrays(1, &_.draw.vao);                                                       GLCHK;
    glBindVertexArray(_.draw.vao);                                                              GLCHK;
    glEnableVertexAttribArray(0);                                                               GLCHK;
    glEnableVertexAttribArray(1);                                                               GLCHK;
    glEnableVertexAttribArray(2);                                                               GLCHK;
    glEnableVertexAttribArray(3);                                                               GLCHK;
    glEnable(GL_DEPTH_TEST);                                                                    GLCHK;
    glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);                                                    GLCHK;
    glPointSize(1);                                                                             GLCHK;
    glClearColor(0.6f, 0.6f, 0.6f, 1.f);                                                        GLCHK;

    // set view and projection
    _.camera.view = glm::lookAt(glm::vec3{0,-4.5,2}, {0,0,-0.3}, {0,0,1});
}

OpenGL::~OpenGL()
{
    delete &_;
}

void* OpenGL::allocate(size_t size) 
{
    // remember the stride in the memory pipeline
    _.memory.stride = size; size *= _.pipeline.items.size();

    // create the buffer
    GLbitfield flags = GL_MAP_READ_BIT | GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT;
    glGenBuffers(1, &_.draw.vbo);                                                               GLCHK;
    glBindBuffer(GL_ARRAY_BUFFER, _.draw.vbo);                                                  GLCHK;
    glBufferStorage(GL_ARRAY_BUFFER, size, nullptr, flags);                                     GLCHK;
    _.memory.ptr = (GLubyte*) glMapBufferRange(GL_ARRAY_BUFFER, 0, size, flags);                GLCHK;
    glBindBuffer(GL_ARRAY_BUFFER, 0);                                                           GLCHK;

    // create the pipeline
    GLbyte* ptr = (GLbyte*)_.memory.ptr; for(int i=0; i<int(_.pipeline.items.size()); ++i, ptr+=_.memory.stride)
        _.pipeline.ptr(i) = ptr;

    // return pointer to first pipeline item
    std::memset(_.memory.ptr, 0xFF, size);
    return _.pipeline.ptr();
}

void* OpenGL::operator()(time dt, iterator begin, iterator end, bool& quit) 
{
    // clear the buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // bind the vao and configure the vertex array pointers -- changes every frame since offset changes every frame
    glBindVertexArray(_.draw.vao);                                                              GLCHK;
    glBindBuffer(GL_ARRAY_BUFFER, _.draw.vbo);                                                  GLCHK;
    glVertexAttribPointer(0, sizeof(begin->pos)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->pos - _.memory.ptr)); GLCHK;
    glVertexAttribPointer(1, sizeof(begin->clr)/sizeof(GLbyte),  GL_UNSIGNED_BYTE,  GL_TRUE,  sizeof(*begin), (GLvoid*)((GLubyte*)&begin->clr - _.memory.ptr)); GLCHK;
    glVertexAttribPointer(2, sizeof(begin->vel)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->vel - _.memory.ptr)); GLCHK;
    glVertexAttribPointer(3, sizeof(begin->age)/sizeof(GLfloat), GL_FLOAT,          GL_FALSE, sizeof(*begin), (GLvoid*)((GLubyte*)&begin->age - _.memory.ptr)); GLCHK;

    // draw the current frame
    auto mvp = _.camera.projection * _.camera.view;
    glUseProgram(_.draw.pgm);                                                                   GLCHK;
    glUniformMatrix4fv(_.draw.mvp, 1, GL_FALSE, glm::value_ptr(mvp));                           GLCHK;
    glDrawArrays(GL_POINTS, 0, GLsizei(end-begin));                                             GLCHK;
    glfwSwapBuffers(_.draw.win);

    // occassionally run the windows event loop -- doing it more often kills performance
    if ((_.camera.time += dt) > 0.5)
        glfwPollEvents();

    // mark the end of the current frame
    auto& cur = _.pipeline.sync(); if (cur) { glDeleteSync(cur);                                GLCHK;}
    cur = glFenceSync(GL_SYNC_GPU_COMMANDS_COMPLETE, 0);                                        GLCHK;

    // advance to next frame
    _.pipeline.advance();

    // wait for the next frame to be free -- wait for up to 1 second
    auto& nxt = _.pipeline.sync(); if (nxt) {
        GLenum rc = glClientWaitSync(nxt, 0, 0); if (GL_ALREADY_SIGNALED != rc) {
            for (size_t i=0; GL_ALREADY_SIGNALED!=rc && GL_TIMEOUT_EXPIRED!=rc && i<1*1000; ++i) {
                rc = glClientWaitSync(nxt, GL_SYNC_FLUSH_COMMANDS_BIT, 1000*1000);              GLCHK;
            }
            fputs("OpenGL::glClientWaitSync(), waited\n", stderr);
        }
    }

    // return the pointer to the next frame
    quit = glfwWindowShouldClose(_.draw.win) > 0;
    return _.pipeline.ptr();
}


}}
