/**********
Copyright 2017 Intel Corporation.

The source code, information and material ('Material') contained herein is owned
 by Intel Corporation or its suppliers or licensors, and title to such Material 
 remains with Intel Corporation or its suppliers or licensors. The Material contains
 proprietary information of Intel or its suppliers and licensors. The Material is 
 protected by worldwide copyright laws and treaty provisions. No part of the Material
 may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, 
 distributed or disclosed in any way without Intel's prior express written permission. 
 No license under any patent, copyright or other intellectual property rights in 
 the Material is granted to or conferred upon you, either expressly, by implication, 
 inducement, estoppel or otherwise. Any license under such intellectual property 
 rights must be express and approved by Intel in writing.

Include any supplier copyright notices as supplier requires Intel to use.

Include supplier trademarks or logos as supplier requires Intel to use, preceded 
by an asterisk. An asterisked footnote can be added as follows: *Third Party 
trademarks are the property of their respective owners.

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice 
or any other notice embedded in Materials by Intel or Intel's suppliers or licensors 
in any way.
**********/
#include "Sources.h"

#include <vector>
#include <chrono>
#include <algorithm>
#include <memory>

namespace Particle { namespace Sources {

/* ----------------------------------------------------------------------------------------------------------------------------- */

struct Flow::Data
{
    time t = 0;
    size_t numCreated = 0;
    double particlesPerSecond = 0.;
    std::vector<std::unique_ptr<Initializer>> initializers;
};

// Constructor
Flow::Flow(double particlesPerSecond,
    std::vector<Initializer*>&& initializers) 
: 
    _(*new Data)
{
    _.particlesPerSecond = particlesPerSecond;
    for(auto&& i : initializers) _.initializers.emplace_back(i);
}

// Destructor
Flow::~Flow()
{
    delete &_;
}

iterator Flow::operator()(time dt, iterator begin, iterator end)
{
    iterator iter = begin;

    // calcuate the number of particles we should have at this time
    _.t += dt; size_t numCreated = size_t(_.particlesPerSecond * _.t);
    
    // if it's more than what we already have
    if (numCreated > _.numCreated) {

        // calculate how many we can create
        numCreated = std::min(numCreated - _.numCreated, size_t(end - begin)); 

        // run all the initializers on the new particles
        #pragma omp parallel for schedule(static, 1024)
        for(int i=0; i<int(numCreated); ++i) // 'i': index variable in OpenMP 'for' statement must have signed integral type
            for (auto& initializer : _.initializers)
                (*initializer)(dt, begin+i);

        // keep track of how many we've created
        _.numCreated += numCreated;

        // return the end of the created particles
        return begin + numCreated;
    } else
        return begin;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */

struct Max::Data
{
    std::vector<std::unique_ptr<Initializer>> initializers;
};

Max::Max(std::vector<Initializer*>&& initializers) : _(*new Data)
{
    for(auto&& i : initializers) _.initializers.emplace_back(i);
}

Max::~Max()
{
    delete &_;
}

iterator Max::operator()(time dt, iterator begin, iterator end)
{
    // create all possible particles 
    size_t cnt = end-begin; 
#pragma omp parallel for schedule(static, 1024)
    for(int i=0; i<int(cnt); ++i) // 'i': index variable in OpenMP 'for' statement must have signed integral type
        for (auto& initializer : _.initializers)
            (*initializer)(dt, begin+i);
    return begin + cnt;
}

}}
